# Copyright (C) 2004 Aaron Bentley
# <aaron.bentley@utoronto.ca>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
import sys
import os
import os.path
import pybaz as arch
import pybaz.util
import commands

__docformat__ = "restructuredtext"
__doc__ = "Support for aba-style external command scripts"

class AbaCmds:
    """
    This class supports external aba-style subcommands in any programming
    language.
    """
    
    abadir = []
    def __init__(self):
        """Determine the path to the aba commands"""
        self.abadir=os.path.expanduser("~/.fai/aba")
        if not os.access(self.abadir, os.X_OK):
            self.abadir="/etc/fai/aba"
        if not os.access(self.abadir, os.X_OK):
            self.abadir=sys.path[0]+"/aba"
        if not os.access(self.abadir, os.X_OK):
            self.abadir=""

    def list_commands(self):
        """
        Display a list of short descriptions of available commands
        """

        if (self.abadir == ""):
            return
        for command in self.get_commands():
            if commands.suggestions.has_key(command.name) or \
                commands.commands.has_key(command.name):
                continue
            command.command_exec(['desc'])

    def get_commands(self):
        """
        Return a sorted list of available aba commands.

        :rtype: a list of `AbaCommand`
        """
        extcommands = []
        if self.abadir != "":
            for k in os.listdir(self.abadir+"/commands"):
                cmd = self.is_command(k)
                if cmd:
                    extcommands.append(cmd)
        extcommands.sort()
        return extcommands

    def is_command(self, command):
        """
        Determine whether a given command is available.

        :param command: The name of the command to look for
        :type command: string
        :rtype: `AbaCommand`
        """
        if (self.abadir == ""):
            return False
        cmd = AbaCommand(command, self.abadir)
        if cmd.exists():
            return cmd
        else:
            return None

class AbaCommand:
    """
    An external, aba-style command script
    """
    def __init__(self, name, abadir):
        self.name = name
        self.path = abadir + "/commands/" + name
        self.abadir = abadir

    def __str__(self):
        return self.name

    def __lt__(self, arg):
        return self.name < arg.name

    def __gt__(self, arg):
        return self.name > arg.name

    def exists(self):
        """
        Determines whether this command really exists.

        Using commands that do not exist is frowned upon.
        """
        return os.access(self.path, os.X_OK) and os.path.isfile(self.path)

    def __call__(self, cmdline):
        try:
            self.run(cmdline.split())
        except Exception, e:
            print e

    def command_exec(self, args, expected=0):
        """
        Invoke a command.

        :param args: The arguments to supply to the command
        :type args: List of strings
        :param expected: Return values to expect
        :type expected: tuple of integers
        :rtype: integer
        """
        if (self.abadir == ""):
            return
        os.environ['abadir'] = self.abadir
        os.environ['abaname'] = 'fai'
        arch.util.exec_safe(self.path, args, stderr=sys.stderr, 
                            stdout=sys.stdout, expected=expected)

    def run(self, args):
        """
        Run the command.

        :param args: The arguments to pass to the command
        :type args: list of strings
        """
        exarg = ['exec']
        exarg.extend(args)
        self.command_exec(exarg, [0, 1, 2])

    def help(self):
        """
        Print command help to stdout.
        """
        self.command_exec(['exec', '-H'])

# arch-tag: c4325683-03f6-4436-bf8a-04e59195f3c1
