#!/usr/bin/env python

#   Copyright (C) 2004 Yannick Gingras <ygingras@ygingras.net>

#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.

#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Add a GNU Arch tagline to files"""

import sys
import os
import time
import re

from email import Utils
from pprint import pprint
from optparse import OptionParser
from tempfile import NamedTemporaryFile, mkstemp

# leading comments are simple strings, comments that need to be closed
# are a tuple of leading and trailing comments
EXT_COMMENTS = {".pl"   : "#",
                ".py"   : "#",
                ".sh"   : "#",
                ".txt"  : "#",
                ".dat"  : "#",
                ".mk"   : "#",
                ".m4"   : "#",
                ".rb"   : "#",
                ".r"    : "#",
                ".R"    : "#",
                ".cpp"  : "//",
                ".hpp"  : "//",
                ".java" : "//",
                ".asm"  : ";;",
                ".lisp" : ";;",
                ".el"   : ";;",
                ".scm"  : ";;",
                ".ada"  : "--",
                ".adb"  : "--",
                ".vhd"  : "--",
                ".vhdl" : "--",
                ".ml"   : ("(*", "*)"),
                ".mli"  : ("(*", "*)"),
                ".mll"  : ("(*", "*)"),
                ".h"    : ("/*", "*/"),
                ".c"    : ("/*", "*/"),
                ".xml"  : ("<!--", "-->"),
                ".htm"  : ("<!--", "-->"),
                ".html" : ("<!--", "-->"),
                ".xhtml": ("<!--", "-->"),
                ".dsl"  : ("<!--", "-->"),
                ".dtml" : ("<!--", "-->"),
                ".php"  : ("<?php /*", "*/ ?>"),
                ".php3" : ("<?php /*", "*/ ?>"),
                }

NOEXT_COMMENTS = {"Makefile" : "#",
                  "README" : "",
                  ".arch-inventory" : "#",
                  }

# still need a reliable source of mimes, try file(1) on this file...
MIME_COMMENTS = {}

TAG = ("arch-", "tag:") # we assemble it live to prevent confusion

SHE_BAM = "#!"

# Here is the plan: if we find more than on match for this one in a
# single file starting with a #!, it's probably some dialect of Lisp
# and the comment char should be ";" and not "#"
LISP_PAT = re.compile(r"^\(def(ine|un|macro|parameter|var|constant|class|method|struct)")

def fail(msg=""):
    msg = msg or "Argh !"
    sys.stderr.write(msg + "\n")
    sys.exit(1)    

def save(filename, data):
    """save data in under filename breaking existing hardlinks if any"""
    # TODO: audit properly for race conditions
    try:
        privs = os.stat(filename)[0]
    except OSError:
        privs = None
    fd, tempname = mkstemp(dir=os.path.dirname(filename))
    file = os.fdopen(fd, "w")
    file.write(data)
    file.flush()
    os.rename(tempname, filename)
    if privs:
        os.chmod(filename, privs)
    file.close()
    
class NoCommentSyntax(Exception):
    def __init__(self, filetype):
        self.filetype = filetype

    def __str__(self):
        return ("** Can't find comment syntax for " + self.filetype  + "\n"
               "** please consider contacting the developer and \n"
               "** telling him how to comment your file.\n")

def isLisp(file):
    """Check if a file contains Lisp.  Don't rely to much on that one,
    it's probably broken."""
    matches = filter(lambda l:l,
                     map(lambda line:LISP_PAT.findall(line),
                         open(file).readlines()))
    return len(matches) >= 2

def getComment(file):
    ext = os.path.splitext(file)[1]
    (dirname, basename) = os.path.split(file)

    if EXT_COMMENTS.has_key(ext):
        return EXT_COMMENTS[ext]
    elif NOEXT_COMMENTS.has_key(basename):
        return NOEXT_COMMENTS[basename]
    elif open(file).read(2) == SHE_BAM:
        # all know languages (ATM) that support the #! have "#" for
        # comments except dialects of Lisp
        if isLisp(file):
            return ";;"
        else:
            return "#"
    else:
        raise NoCommentSyntax(ext or basename)


def genId(file, opt):
    if opt.tltl:
        (dirname, basename) = os.path.split(file)
        pipe = os.popen("tla my-id 2>/dev/null")
        name = pipe.read()
        try:
            name = name[:name.index("<")].strip()
        except ValueError:
            pass            
        tstamp = time.strftime("%Y-%m-%d %X")
        if not pipe.close():
            return " ".join([name, tstamp, ("(%s)" % basename)])
    pipe = os.popen("uuidgen 2>/dev/null")
    id = pipe.read().strip()
    if pipe.close():
        id = Utils.make_msgid()[1:-1] # don't keep "<" and ">"
    return id


class AlreadyTagged(Exception):
    pass

def isTagged(data):
    "can we find something that looks like a tag in there ?"
    # this is a bit ugly, there are a few way to assemble the TAG list
    # that yielld a tag...    
    if data.find(TAG[-1]) != -1:
        return True

    # Right now this will be matched by the 1st try but the proper way
    # to check is to get the tagging method 1st.
    # TODO: Get the tagging method 1st
    if data.find("".join(TAG)) != -1:
        return True

    return False


def tagFile(file, opt):
    data = open(file).read()
    
    # don't double tag
    if isTagged(data):
        raise AlreadyTagged()      

    comment = getComment(file)
    if opt.implicit:
        tag = TAG[-1]
    else:
        tag = "".join(TAG)
        
    id = opt.id or genId(file, opt)
    if type(comment) in (type(""), type(unicode(""))):
        tagline = " ".join((comment, tag, id))
    else:
        tagline = " ".join((comment[0], tag, id, comment[1]))

    data = data + "\n" + tagline + "\n"
    save(file, data)


# TODO: test something more representative
def selfTest(instPath):
    exts = EXT_COMMENTS.keys()
    exts.sort()
    for ext in exts:
        file = NamedTemporaryFile(suffix=ext)
        if os.system(instPath + " " + file.name):
            fail("self-test failure, check your installation")
        if open(file.name).read().find("".join(TAG)) == -1:
            fail("self-test failed silently (thats bad), "
                 "check your installation")
    print "all tests succeeded !"
    

def parseOptions():
    usage = "usage: %prog [options] FILE1 [FILE2 ...]"
    desc  = "Add a GNU Arch tagline to files" 
    parser = OptionParser(usage=usage, description=desc)
    parser.add_option("-t", "--test",
                      action="store_true", dest="test", default=False,
                      help="Run a self-test.")
    parser.add_option("-T", "--tltl",
                      action="store_true", dest="tltl", default=False,
                      help="Make a Tom Lord style Tag Line.")
    parser.add_option("-s", "--sensitive",
                      action="store_true", dest="sensitive", default=False,
                      help=("Stop tagging if one file can't be tagged.  "
                            "The default is to keep going with the other"
                            " files."))
    parser.add_option("-i", "--id",
                      dest="id", default=None,
                      help="Specify ID, instead of using auto-generated id.",
                      metavar="ID")
    parser.add_option("-I", "--implicit",
                      action="store_true", dest="implicit", default=False,
                      help="Use implicit tags (deprecated).")

    return parser.parse_args(sys.argv)
    

if __name__ == "__main__":
    option, args = parseOptions()
    if option.test:
        selfTest(args[0])

    taggedFiles = []
    unsupportedFiles = []
    errors = []
    
    for file in args[1:]:
        try:
            tagFile(file, option)
        except AlreadyTagged, e:
            taggedFiles.append(file)
            if option.sensitive:
                fail(file + " is already tagged.")
        except NoCommentSyntax, e:
            unsupportedFiles.append(file)
            if option.sensitive:
                fail(str(e))
        except Exception, e:
            errors.append(str(e))
            if option.sensitive:
                fail(str(e))

    if taggedFiles:
        print "The following files were already tagged:",
        print " ".join(taggedFiles)

    if unsupportedFiles:
        print "The following files were unsupported:",
        print " ".join(unsupportedFiles)
        print "If you know how to tag them, please inform the author."

    if errors:
        print "The following unexpexted errors occured:"
        print "", "\n ".join(errors)
        
            
# arch-tag: 9806bff2-6612-44ec-8817-61d365b5ce77
