#!/usr/bin/env python
# arch-tag: 6e2d3c18-a7b2-4d3c-a4ae-0fb1773e7214
# Copyright (C) 2003 Ben Burns <bburns@mailops.com>
#               2004 David Allouche <david@allouche.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Main test suite
"""

import types
import sys
import os
import shutil
import arch
from arch import errors
from arch.pathname import DirName, FileName, PathName
import framework
from framework import TestCase

# import arch.backends.logger
# arch.backend.logger = backends.logger.DebugLogger()

class ArchBasic(TestCase):

    tests = []

    def default_my_id_empty(self):
        """my_id returns None when unset."""
        my_id = arch.my_id()
        self.failUnlessEqual(None, my_id)
    tests.append('default_my_id_empty')

    def default_archive_empty(self):
        """archives is empty when no archive is registered."""
        archives = list(arch.iter_archives())
        self.failUnlessEqual(len(archives), 0)
    tests.append('default_archive_empty')

    def set_my_id(self):
        """set_my_id effects arch.my_id()."""
        arch.set_my_id(self.my_id)
        my_id = arch.my_id()
        self.failUnlessEqual(my_id, self.my_id)
    tests.append('set_my_id')

    def set_invalid_myid(self):
        """set_my_id fails for invalid id and preserves current id."""
        arch.set_my_id(self.my_id)
        try:
            arch.set_my_id("not a valid email address")
        except arch.ExecProblem:
            my_id = arch.my_id()
            self.failUnlessEqual(my_id, self.my_id)
        else:
            self.fail("invalid email address was accepted")
    tests.append('set_invalid_myid')

    def make_archive(self):
        """make_archive creates and registers archive at location."""
        name = self.arch_name
        location = self.arch_dir/name
        newArchive = arch.make_archive(name, location)
        self.failUnlessEqual(newArchive.name, name)
        self.failUnlessEqual(newArchive.location, location)
        archives = list(arch.iter_archives())
        self.failUnlessEqual(len(archives), 1)
        archive = archives[0]
        self.failUnlessEqual(archive.name, name)
        self.failUnless(os.path.samefile(archive.location, location))
        return archive
    tests.append('make_archive')


class Archive(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()

    tests = []

    def archive_location_readonly(self):
        """Archive.location is present and read-only."""
        self.failUnless(hasattr(self.archive, 'location'))
        def setArchiveLocation(): self.archive.location = 'error'
        self.failUnlessRaises(AttributeError, setArchiveLocation)
    tests.append('archive_location_readonly')

    def archive_name_readonly(self):
        """Archive.name is present and read-only."""
        self.failUnless(hasattr(self.archive, 'name'))
        def setArchiveName(): self.archive.name = 'error'
        self.failUnlessRaises(AttributeError, setArchiveName)
    tests.append('archive_name_readonly')

    def make_already_made_archive(self):
        """Making the same archive twice fails."""
        try: arch.make_archive(self.archive.name, self.archive.location)
        except arch.ExecProblem: pass
        else: self.fail("could create same archive twice")
    tests.append('make_already_made_archive')

    def archive_setup(self):
        """Version.setup works."""
        arch_name = self.arch_name
        archive = self.archive
        version_id = arch_name + "/example--test--0.1"
        version = arch.Version(version_id)
        version.setup()
        cats = list(archive.iter_categories())
        self.failUnlessEqual(len(cats), 1)
        self.failUnlessEqual(cats[0].fullname, arch_name + '/example')
        branches = list(cats[0].iter_branches())
        self.failUnlessEqual(len(branches), 1)
        self.failUnlessEqual(branches[0].fullname,
                             arch_name + '/example--test')
        versions = list(branches[0].iter_versions())
        self.failUnlessEqual(len(versions), 1)
        self.failUnlessEqual(versions[0].fullname, version_id)
    tests.append('archive_setup')

    def archive_setup_nameless_branch(self):
        """Version.setup (nameless branch) works."""
        arch_name = self.arch_name
        archive = self.archive
        version_id = arch_name + "/example--0.1"
        version = arch.Version(version_id)
        version.setup()
        cats = list(archive.iter_categories())
        self.failUnlessEqual(len(cats), 1)
        self.failUnlessEqual(cats[0].fullname, arch_name + '/example')
        branches = list(cats[0].iter_branches())
        self.failUnlessEqual(len(branches), 1)
        self.failUnlessEqual(list(cats[0].iter_branches())[0].fullname,
                             arch_name + '/example')
        versions = list(branches[0].iter_versions())
        self.failUnlessEqual(len(versions), 1)
        self.failUnlessEqual(versions[0].fullname, version_id)
    tests.append('archive_setup_nameless_branch')

    def _help_make_mirror(self, kwargs, meta_name, meta_expected):
        master = self.archive
        mirror_name = master.name + '-MIRROR'
        location = self.arch_dir/mirror_name
        mirror = master.make_mirror(mirror_name, location, **kwargs)
        meta_file = DirName(mirror.location)/'=meta-info'/meta_name
        self.failUnless(os.path.isfile(meta_file))
        meta_data = open(meta_file).read()
        self.failUnlessEqual(meta_expected, meta_data)

    def make_mirror(self):
        """Archive.make_mirror works for creating a -MIRROR."""
        self._help_make_mirror({}, 'mirror', self.arch_name + '\n')
    tests.append('make_mirror')

    def make_mirror_listing(self):
        """Archive.make_mirror can set up http-blows."""
        self._help_make_mirror({'listing': True}, 'http-blows',
                               'it sure does\n')
    tests.append('make_mirror_listing')

    def make_mirror_signed(self):
        """Archive.make_mirror can set up signed-archive."""
        self._help_make_mirror({'signed': True}, 'signed-archive',
                               'system cracking is (nearly always) lame\n')
    tests.append('make_mirror_signed')

    def meta_info(self):
        """Archive._meta_info works."""
        self.assertEqual(self.archive.name, self.archive._meta_info('name'))
    tests.append('meta_info')

    def open_meta_info(self, name, flags):
        return open(DirName(self.archive.location)/'=meta-info'/name, flags)

    def meta_info_missing(self):
        """arch.tla_tells_empty_meta_info is set correctly."""
        if arch._builtin.tla_tells_empty_meta_info:
            try:
                self.archive._meta_info('not_here')
                self.fail("arch.tla_tells_empty_meta_info is True, but"
                          " archive-meta-info exit(0) on missing meta-info.")
            except KeyError:
                pass
        else:
            try:
                self.assertEqual(None, self.archive._meta_info('not_here'))
            except KeyError:
                self.fail("arch.tla_tells_empty_meta_info is False, but"
                          " archive-meta-info exit(1) on missing meta-info.")
        self.open_meta_info('empty', 'w')
        self.assertEqual(None, self.archive._meta_info('empty'))
        print >> self.open_meta_info('empty-line', 'w')
        self.assertEqual(str(), self.archive._meta_info('empty-line'))
    tests.append('meta_info_missing')

    def has_meta_info(self):
        """Archive._has_meta_info works."""
        self.failIf(self.archive._has_meta_info('not_here'))
        self.open_meta_info('empty', 'w')
        if arch._builtin.tla_tells_empty_meta_info:
            self.assert_(self.archive._has_meta_info('empty'))
        else:
            self.failIf(self.archive._has_meta_info('empty'))
        print >> self.open_meta_info('empty-line', 'w')
        self.assert_(self.archive._has_meta_info('empty-line'))
        print >> self.open_meta_info('something', 'w'), 'hello'
        self.assert_(self.archive._has_meta_info('something'))
    tests.append('has_meta_info')

    def is_signed(self):
        """Archive.is_signed works."""
        mirror_name = self.archive.name + '-MIRROR'
        location = self.arch_dir/mirror_name
        mirror = self.archive.make_mirror(mirror_name, location, signed=True)
        self.assertEqual(False, self.archive.is_signed)
        self.assertEqual(True, mirror.is_signed)
    tests.append('is_signed')

    def has_listings(self):
        """Archive.has_listings works."""
        mirror_name = self.archive.name + '-MIRROR'
        location = self.arch_dir/mirror_name
        mirror = self.archive.make_mirror(mirror_name, location, listing=True)
        self.assertEqual(False, self.archive.has_listings)
        self.assertEqual(True, mirror.has_listings)
    tests.append('has_listings')

    def open_version(self, flags):
        return open(DirName(self.archive.location)/'.archive-version', flags)

    def version_string(self):
        """Archive.version_string works."""
        version_file = self.open_version('r')
        version_string = version_file.read().rstrip('\n')
        version_file.close()
        self.assertEqual(version_string, self.archive.version_string)
    tests.append('version_string')


class MirrorArchive(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive_and_mirror()

    tests = []

    def is_mirror(self):
        """Archive.is_mirror works."""
        self.assert_(not self.archive.is_mirror)
        self.assert_(self.mirror.is_mirror)
    tests.append('is_mirror')

    def official_name(self):
        """Archive.official_name works."""
        self.assertEqual(self.archive.official_name, self.archive.name)
        self.assertEqual(self.mirror.official_name, self.archive.name)
    tests.append('official_name')

    def mirror_official_name(self):
        """Archive.make mirror use official name."""
        mirror2_name = self.mirror.name + '-2'
        location2 = self.arch_dir/mirror2_name
        mirror2 = self.mirror.make_mirror(mirror2_name, location2)
        self.assertEqual(mirror2.official_name, self.archive.name)
    tests.append('mirror_official_name')


class InventoryTree(TestCase):

    """Test cases that work on a tree but need no associated archive"""

    def extraSetup(self):
        self.set_my_id()
        self.create_working_tree()

    tests = []

    def inventory(self):
        """Basic inventory."""
        wt = self.working_tree
        open(wt/'ahoy', 'w').write('')
        wt.add_tag(wt/'ahoy')
        os.mkdir(wt/'bar')
        wt.add_tag(wt/'bar')
        open(wt/'bar'/'baz', 'w').write('')
        wt.add_tag(wt/'bar'/'baz')
        inv = list(wt.iter_inventory(source=True, both=True))
        result = map(type, inv)
        expected = [FileName, DirName, FileName]
        self.assertEqual(expected, result)
        result = map(str, inv)
        expected = ['ahoy', 'bar', 'bar/baz']
        self.assertEqual(expected, result)
        inv = list(wt.iter_inventory(source=True, files=True))
        self.assertEqual([FileName, FileName], map(type, inv))
        self.assertEqual(['ahoy', 'bar/baz'], map(str, inv))
        inv = list(wt.iter_inventory(source=True, directories=True))
        self.assertEqual([DirName], map(type, inv))
        self.assertEqual(['bar'], map(str, inv))
    tests.append('inventory')

    def inventory_ids(self):
        """Inventory with ids"""
        wt = self.working_tree
        wt.tagging_method = 'names'
        open(wt/'ahoy', 'w').write('')
        os.mkdir(wt/'bar')
        open(wt/'bar'/'baz', 'w').write('')
        inv = list(wt.iter_inventory_ids(source=True, both=True))
        result = map(lambda L: type(L[1]), inv)
        expected = [FileName, DirName, FileName]
        self.assertEqual(expected, result)
        result = map(lambda L: str(L[1]), inv)
        expected = ['ahoy', 'bar', 'bar/baz']
        self.assertEqual(expected, result)
        result = map(lambda L: L[0], inv)
        expected = ['?./ahoy', '?./bar', '?./bar/baz']
        self.assertEqual(expected, result)
        inv = list(wt.iter_inventory_ids(source=True, files=True))
        self.assertEqual([FileName, FileName], map(lambda L: type(L[1]), inv))
        self.assertEqual(['ahoy', 'bar/baz'], map(lambda L: str(L[1]), inv))
        self.assertEqual(['?./ahoy', '?./bar/baz'], map(lambda L: L[0], inv))
        inv = list(wt.iter_inventory_ids(source=True, directories=True))
        self.assertEqual([DirName], map(lambda L: type(L[1]), inv))
        self.assertEqual(['bar'], map(lambda L: str(L[1]), inv))
        self.assertEqual(['?./bar'], map(lambda L: L[0], inv))
    tests.append('inventory_ids')

    def inventory_names(self):
        """iter_inventory names keyword overrides untagged-source setting."""
        wt = self.working_tree
        print >> open(wt/'{arch}'/'=tagging-method', 'w'),\
              '\n'.join(('explicit', 'untagged-source precious',''))
        foo = FileName('foo')
        print >> open(wt/foo, 'w')
        result = list(wt.iter_inventory(precious=True, files=True))
        self.assertEqual([foo], result) # foo is untagged-source precious
        result = list(wt.iter_inventory(source=True, files=True))
        self.assertEqual([], result) # thus foo is not source
        result = list(wt.iter_inventory(source=True, files=True, names=True))
        self.assertEqual([foo], result) # however it has a source name
    tests.append('inventory_names')

    def get_tag(self):
        """ArchSourceTree.get_tag works."""
        wt = self.working_tree
        wt.tagging_method = 'tagline'
        foo = FileName('foo')
        print >> open(wt/foo, 'w')
        self.assertEqual(None, wt.get_tag(foo))
        print >> open(wt/foo, 'w'), 'arch-tag: yadda'
        self.assertEqual('i_yadda', wt.get_tag(foo))
        wt.add_tag(foo)
        print >> open(wt/'.arch-ids'/foo+'.id', 'w'), 'yadda'
        self.assertEqual('x_yadda', wt.get_tag(foo))
    tests.append('get_tag')

    def inventory_file_space(self):
        """iter_inventory handles spaces in file and dir names correctly."""
        if tla_skip_whitespace_tests(): return
        wt = self.working_tree
        filename = FileName('file name')
        dirname = DirName('dir name')
        open(wt/filename, 'w').write('')
        wt.add_tag(filename)
        os.mkdir(wt/dirname)
        wt.add_tag(dirname)
        expected = [filename]
        result = list(wt.iter_inventory(source=True, files=True))
        self.assertEqual(expected, result) # inventory -s
        expected = [dirname]
        result = list(wt.iter_inventory(source=True, directories=True))
        self.assertEqual(expected, result) # inventory -s -d
        expected = [filename, dirname]
        expected.sort()
        result = list(wt.iter_inventory(source=True, both=True))
        result.sort()
        self.assertEqual(expected, result) # inventory -s -B
        nestedname = 'nested tree'
        os.mkdir(wt/nestedname)
        nested = arch.init_tree(wt/nestedname, nested=True)
        expected = [nested]
        result = list(wt.iter_inventory(trees=True))
        self.assertEqual(expected, result) # inventory -t
    tests.append('inventory_file_space')

    def inventory_limit(self):
        """Inventory with a limit"""
        wt = self.working_tree
        open(wt/'ahoy', 'w').write('')
        wt.add_tag(wt/'ahoy')
        os.mkdir(wt/'bar')
        wt.add_tag(wt/'bar')
        open(wt/'bar'/'baz', 'w').write('')
        wt.add_tag(wt/'bar'/'baz')
        inv = list(wt.iter_inventory(limit='bar', source=True, both=True))
        self.assertEqual([FileName], map(type, inv))
        self.assertEqual(['bar/baz'], map(str, inv))
        def thunk():
            return wt.iter_inventory(limit=('a', 'b'), source=True, both=True)
        self.assertRaises(TypeError, thunk)
        def thunk():
            return wt.iter_inventory(limit='/tmp', source=True, both=True)
        self.assertRaises(ValueError, thunk)
    tests.append('inventory_limit')

    def inventory_ids_limit(self):
        """Inventory with ids"""
        wt = self.working_tree
        wt.tagging_method = 'names'
        open(wt/'ahoy', 'w').write('')
        os.mkdir(wt/'bar')
        open(wt/'bar'/'baz', 'w').write('')
        inv = list(wt.iter_inventory_ids(limit='bar', source=True, both=True))
        self.assertEqual([FileName], map(lambda L: type(L[1]), inv))
        self.assertEqual(['bar/baz'], map(lambda L: str(L[1]), inv))
        expected, result = ['?bar/baz'], map(lambda L: L[0], inv)
        self.assertEqual(expected, result) # XXX will fail when tla is fixed
        def thunk():
            return wt.iter_inventory_ids(
                limit=('a', 'b'), source=True, both=True)
        self.assertRaises(TypeError, thunk)
        def thunk():
            return wt.iter_inventory_ids(limit='/tmp', source=True, both=True)
        self.assertRaises(ValueError, thunk)
    tests.append('inventory_ids_limit')


class InitTree(TestCase):

    tests = []

    def init_tree(self):
        """init_tree works (w/o version) for unnested and nested trees."""
        os.mkdir(self.working_dir)
        arch.init_tree(self.working_dir)
        os.mkdir(self.nested_dir)
        arch.init_tree(self.nested_dir, nested=True)
    tests.append('init_tree')

    def init_tree_not_nested(self):
        """init_tree (w/o version) fails for nested if "nested" is not set."""
        os.mkdir(self.working_dir)
        arch.init_tree(self.working_dir)
        os.mkdir(self.nested_dir)
        self.failUnlessRaises(errors.ExecProblem,
                              lambda: arch.init_tree(self.nested_dir))
    tests.append('init_tree_not_nested')

    def init_tree_with_version(self):
        """init_tree with version set tree-version and create log-version."""
        os.mkdir(self.working_dir)
        wt = arch.init_tree(self.working_dir, self.version)
        self.failUnlessEqual(self.version, wt.tree_version)
        self.failUnlessEqual(1, len(list(wt.iter_log_versions())))
        self.failUnlessEqual(self.version, wt.iter_log_versions().next())
    tests.append('init_tree_with_version')

    def tree_root(self):
        """tree_root works"""
        os.mkdir(self.working_dir)
        tree = arch.init_tree(self.working_dir)
        subdir = tree/'subdir'
        os.mkdir(subdir)
        subdir_tree = arch.tree_root(subdir)
        self.assertEqual(tree, subdir_tree)
    tests.append('tree_root')


class WorkingTree(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_working_tree()

    tests = []

    def log_versions(self):
        """Log version addition, iteration and deletion."""
        wt = self.working_tree
        self.failUnlessEqual(0, len(list(wt.iter_log_versions())))
        wt.add_log_version(self.version)
        self.failUnlessEqual(1, len(list(wt.iter_log_versions())))
        self.failUnless(isinstance(wt.iter_log_versions().next(),
                                   arch.Version))
        self.failUnlessEqual(self.version, wt.iter_log_versions().next())
        wt.remove_log_version(self.version)
        self.failUnlessEqual(0, len(list(wt.iter_log_versions())))
    tests.append('log_versions')

    def log_versions_limit(self):
        """Log version listing with a limit."""
        wt = self.working_tree
        vsn = self.version
        jdoe_cat_brn_10 = vsn
        other_arch = arch.Version('alice@example/%s' % vsn.nonarch)
        other_cat = vsn.archive['dog']['brn']['1.0']
        other_brn = vsn.category['trk']['1.0']
        other_vsn = vsn.branch['2']
        all_versions = [vsn, other_arch, other_cat, other_brn, other_vsn]
        for V in all_versions: wt.add_log_version(V)
        all_versions.sort()
        expected = map(str, all_versions)
        expected.sort()
        result = map(str, wt.iter_log_versions())
        result.sort()
        self.assertEqual(expected, result)
        result = list(wt.iter_log_versions(other_arch.archive))
        self.assertEqual([other_arch], result)
        result = map(str, wt.iter_log_versions(vsn.archive))
        result.sort()
        expected = map(str, (vsn, other_cat, other_brn, other_vsn))
        expected.sort()
        self.assertEqual(expected, result)
        result = list(wt.iter_log_versions(other_cat.category))
        self.assertEqual([other_cat], result)
        result = map(str, wt.iter_log_versions(vsn.category))
        result.sort()
        expected = map(str, (vsn, other_brn, other_vsn))
        expected.sort()
        self.assertEqual(expected, result)
        result = list(wt.iter_log_versions(other_brn.branch))
        self.assertEqual([other_brn], result)
        result = map(str, wt.iter_log_versions(vsn.branch))
        result.sort()
        expected = map(str, (vsn, other_vsn))
        expected.sort()
        self.assertEqual(expected, result)
        result = list(wt.iter_log_versions(vsn))
        self.assertEqual([vsn], result)
    tests.append('log_versions_limit')

    def tree_version(self):
        """Setting and getting WorkingTree.version."""
        wt = self.working_tree
        self.failUnlessRaises(arch.errors.TreeVersionError,
                              lambda: wt.tree_version)
        try: wt.tree_version
        except arch.errors.TreeVersionError, E:
            self.assertEqual(E.bad_version, None)
        open(wt/'{arch}'/'++default-version', 'w').write('fooo!!!\n')
        try: wt.tree_version
        except arch.errors.TreeVersionError, E:
            self.assertEqual(E.bad_version, 'fooo!!!')
        self.failUnlessRaises(arch.errors.TreeVersionError,
                              lambda: wt.tree_version)
        wt.tree_version = self.version
        self.failUnless(isinstance(wt.tree_version, arch.Version))
        self.failUnlessEqual(self.version, wt.tree_version)
    tests.append('tree_version')


class TreeRevision(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_version()

    tests = []

    def tree_revision(self):
        """tree_revision works"""
        os.mkdir(self.working_dir)
        wt = arch.init_tree(self.working_dir, self.version)
        wt.import_()
        self.assertEqual(self.version['base-0'], wt.tree_revision)
        m = wt.log_message()
        m['Summary'] = m.description = 'first revision'
        wt.commit(m)
        self.assertEqual(self.version['patch-1'], wt.tree_revision)
    tests.append('tree_revision')


class LogMessage(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_working_tree(self.version)
        nonarch = self.version.nonarch
        archname = self.version.archive.name
        self.file_name = '++log.%s--%s' % (nonarch, archname)

    tests = []

    def create(self):
        """LogMessage creation."""
        wt = self.working_tree
        # message is None if not created
        m = wt.log_message(create=False)
        self.failUnlessEqual(m, None)
        # message is created by default
        m = wt.log_message()
        self.failIfEqual(m, None)
    tests.append('create')

    def name(self):
        """LogMessage.name attribute is read-only."""
        m = self.working_tree.log_message()
        self.assertEqual(m.name, self.working_tree.abspath()/m.name)
        def setLogName(): m.name = 'error'
        self.failUnlessRaises(AttributeError, setLogName)
    tests.append('name')

    def template(self):
        """LogMessage template has expected value and can be parsed."""
        m = self.working_tree.log_message()
        template = '\n'.join(('Summary: ', 'Keywords: ', '', ''))
        self.failUnlessEqual(file(m.name).read(), template)
        self.failUnlessEqual(m['Summary'], '')
        self.failUnlessEqual(m['Keywords'], '')
        self.failUnlessEqual(m.description, '')
    tests.append('template')

    def setup_message(self):
        self.description = '\n'.join(("Hello,  ", "  World!"))
        # leading space in summary dropped, trailing preserved
        self.summary = "Short summary  "
        self.message = '\n'.join(('Summary: ' + self.summary,
                                  'Keywords: ', '', self.description))
        m = self.message
        m = self.working_tree.log_message()
        m.description = self.description
        m['Summary'] = self.summary
        return m

    def check_message_file(self, m):
        self.failUnlessEqual(file(m.name).read(), self.message)

    def check_message_object(self, m):
        self.failUnlessEqual(m['Summary'], self.summary)
        self.failUnlessEqual(m['Keywords'], '')
        self.failUnlessEqual(m.description, self.description)

    def save_load(self):
        """LogMessage save() and load() behave consistently."""
        m = self.setup_message()
        self.check_message_object(m)
        m.save()
        self.check_message_file(m)
        m.load()
        self.check_message_object(m)
    tests.append('save_load')

    def recreate(self):
        """Recreating LogMessage does not overwrite existing message."""
        m = self.setup_message()
        m.save()
        del m
        m = self.working_tree.log_message()
        self.check_message_file(m)
        self.check_message_object(m)
    tests.append('recreate')

    def reload_(self):
        """Reloading LogMessage discards unsaved changes."""
        m = self.setup_message()
        m.save()
        m['Summary'] = 'foo'
        m.load()
        self.check_message_file(m)
        self.check_message_object(m)
    tests.append('reload_')

    def clear(self):
        """LogMessage.clear() method."""
        m = self.setup_message()
        m.clear()
        self.failUnlessEqual(m['Summary'], None)
        self.failUnlessEqual(m.description, None)
        m.save()
        self.failUnlessEqual(file(m.name).read(), '\n')
    tests.append('clear')


class LogMessageWhitespace(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()

    tests = []

    def log_message_whitespace(self):
        """LogMessage files created in a path containing a space."""
        os.mkdir(self.working_dir_newline)
        wt = arch.init_tree(self.working_dir_newline, self.version)
        m = wt.log_message()
        template = '\n'.join(('Summary: ', 'Keywords: ', '', ''))
        self.failUnlessEqual(file(m.name).read(), template)
    tests.append('log_message_whitespace')


class Import(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_version()
        self.create_working_tree(self.version)

    tests = []

    def import_without_log(self):
        """WorkingTree.import w/o log message produces expected message."""
        wt = self.working_tree
        wt.import_()
        logs = tuple(wt.iter_logs())
        self.failUnlessEqual(1, len(logs))
        log = logs[0]
        self.failUnlessEqual(log.revision, self.version['base-0'])
        self.failUnlessEqual(log.summary, 'initial import')
        self.failUnlessEqual(log.description,
                             '\n(automatically generated log message)\n')
    tests.append('import_without_log')

    def _import_with_log_helper(self, m, import_):
        wt = arch.WorkingTree(self.working_dir)
        s = "empty initial import"
        m['Summary'] = s
        d = "Empty initial import.\n"
        m.description = d
        import_()
        logs = tuple(wt.iter_logs())
        self.failUnlessEqual(1, len(logs))
        log = logs[0]
        self.failUnlessEqual(log.revision, self.version['base-0'])
        self.failUnlessEqual(log.summary, s)
        self.failUnlessEqual(log.description, d)

    def import_with_log(self):
        """WorkingTree.import with default log message file name."""
        wt = self.working_tree
        m = wt.log_message()
        def import_():
            m.save()
            wt.import_()
        self._import_with_log_helper(m, import_)
    tests.append('import_with_log')

    def import_with_custom_log(self):
        """WorkingTree.import with specified log message file name."""
        wt = self.working_tree
        m = arch.LogMessage(wt/'+custom_log_message')
        m.clear()
        def import_():
            wt.import_(m)
        self._import_with_log_helper(m, import_)
    tests.append('import_with_custom_log')

    def import_not_empty(self):
        """WorkingTree.import fails on unrecognized and adds source."""
        wt = self.working_tree
        # unrecognized file
        print >> file(wt/'!foo', 'w'), 'bar'
        def import_default(): wt.import_()
        self.failUnlessRaises(errors.ExecProblem, import_default)
        os.remove(wt/'!foo')
        # source file
        print >> file(wt/'source', 'w'), 'hello'
        wt.add_tag('source')
        wt.import_()
        log = wt.iter_logs().next()
        self.failUnlessEqual(list(log.new_files), [FileName('source')])
    tests.append('import_not_empty')


class Commit(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_version()
        self.create_working_tree(self.version)

    tests = []

    def commit_seal_fix(self):
        """Commit, seal, fix, create expected patchlevels."""
        wt = self.working_tree
        wt.import_()
        # commit w/o log message
        self.failUnlessRaises(errors.ExecProblem, lambda: wt.commit())
        # commit revision
        m = wt.log_message()
        m['Summary'] = m.description = 'first revision'
        wt.commit(m)
        self.failUnlessEqual(map(lambda(x): x.revision, wt.iter_logs()),
                             map(lambda(x): self.version[x],
                                 ('base-0', 'patch-1')))
        # fix non-sealed, then seal
        m['Summary'] = m.description = 'second revision'
        self.failUnlessRaises(errors.ExecProblem,
                              lambda: wt.commit(fix=True, log=m))
        wt.commit(seal=True, log=m)
        self.failUnlessEqual(map(lambda(x): x.revision, wt.iter_logs()),
                             map(lambda(x): self.version[x],
                                 ('base-0', 'patch-1', 'version-0')))
        # commit sealed, then fix
        m['Summary'] = m.description = 'third revision'
        self.failUnlessRaises(errors.ExecProblem,
                              lambda: wt.commit(m))
        wt.commit(fix=True, log=m)
        self.failUnlessEqual(map(lambda(x): x.revision, wt.iter_logs()),
                             map(lambda(x): self.version[x],
                                 ('base-0', 'patch-1',
                                  'version-0', 'versionfix-1')))
        # sanity catches trying to fix and seal at the same time
        m['Summary'] = m.description = 'error'
        self.failUnlessRaises(AssertionError,
                              lambda: wt.commit(seal=True, fix=True, log=m))
    tests.append('commit_seal_fix')

    def commit_relative_log(self):
        """Commit with a log file given as a relative path."""
        wt = self.working_tree
        wt.import_()
        tree_dir, tree_name = map(DirName, os.path.split(wt))
        os.chdir(tree_dir)
        log_name = tree_name/'+log'
        f = open(log_name, 'w')
        summary = 'first rev'
        f.write('Summary: %s\n\n' % summary)
        f.close()
        wt.commit(log=log_name)
        def lastlog(): return wt.iter_logs(reverse=True).next()
        self.failUnlessEqual(lastlog().summary, summary)
    tests.append('commit_relative_log')

    def commit_files(self):
        """Commit (maybe with limit) and Patchlog are consistent."""
        wt = self.working_tree
        # import two files
        print >> file(wt/'f1', 'w'), 'hello'
        print >> file(wt/'f2', 'w'), 'world'
        map(wt.add_tag, ('f1', 'f2'))
        wt.import_()
        def lastlog(): return wt.iter_logs(reverse=True).next()
        self.failUnlessEqual(lastlog().new_files, map(FileName, ('f1','f2')))
        # modify both files
        print >> file(wt/'f1', 'w'), 'Hello,'
        print >> file(wt/'f2', 'w'), 'World!'
        # sanity catches non-null but empty file-list
        m = wt.log_message()
        m['Summary'] = m.description = 'first revision'
        def commit_empty(): wt.commit(log=m, file_list=[].__iter__())
        self.failUnlessRaises(AssertionError, commit_empty)
        # commit one file, absolute path
        wt.commit(m, file_list=[wt/'f2'].__iter__())
        self.failUnlessEqual(lastlog().modified_files, [FileName('f2')])
        # commit the other file, relative path
        m['Summary'] = m.description = 'third revision'
        wt.commit(m, file_list=['f1'])
        self.failUnlessEqual(lastlog().modified_files, [FileName('f1')])
    tests.append('commit_files')

    def commit_out_of_date(self):
        """Committing out-of-date fails unless forced."""
        wt = self.working_tree
        wt.import_()
        m = wt.log_message()
        m['Summary'] = m.description = 'first revision'
        wt.commit(m)
        shutil.rmtree(self.working_dir)
        wt = self.version['base-0'].get(self.working_dir)
        m['Summary'] = m.description = 'out of date revision'
        # trying to commit an out-of-date tree fails
        self.failUnlessRaises(errors.ExecProblem, lambda: wt.commit(m))
        # unless the right option is set
        wt.commit(m, out_of_date_ok=True)
        self.failUnlessEqual(map(lambda(x): x.revision, wt.iter_logs()),
                             map(lambda(x): self.version[x],
                                 ('base-0', 'patch-2')))
    tests.append('commit_out_of_date')

    def commit_version(self):
        """Commit on a specified version."""
        wt = self.working_tree
        wt.import_()
        def think(**kw):
            def thunk():
                m = wt.log_message()
                m['Summary'] = m.description = 'first revision'
                wt.commit(m, **kw)
            return thunk
        wt.tree_version = self.other_version
        self.assertRaises(errors.ExecProblem, think())
        think(version=self.version)()
        wt.tree_version = self.version
        self.failUnlessEqual(map(lambda(x): x.revision, wt.iter_logs()),
                             map(lambda(x): self.version[x],
                                 ('base-0', 'patch-1')))
    tests.append('commit_version')

    def iter_commit(self):
        """iter_commit works"""
        wt = self.working_tree
        wt.tagging_method = 'names'
        wt.import_()
        print >> file(wt/'f1', 'w'), 'hello'
        m = wt.log_message()
        m['Summary'] = m.description = 'first revision'
        lines = wt.iter_commit(m)
        changes = [L for L in lines if isinstance(L, arch.TreeChange)]
        self.assertEqual(1, len(changes))
        self.assertEqual(arch.FileAddition, type(changes[0]))
        self.assertEqual('f1', changes[0].name)
    tests.append('iter_commit')

    def iter_commit_files(self):
        """Commit with limit print the right path."""
        wt = self.working_tree
        print >> file(wt/'f1', 'w'), 'hello'
        print >> file(wt/'f2', 'w'), 'world'
        map(wt.add_tag, ('f1', 'f2'))
        wt.import_()
        print >> file(wt/'f1', 'w'), 'Hello,'
        print >> file(wt/'f2', 'w'), 'World!'
        m = wt.log_message()
        m['Summary'] = m.description = 'first revision'
        lines = wt.iter_commit(m, file_list=['f2'].__iter__())
        changes = [L for L in lines if isinstance(L, arch.TreeChange)]
        self.assertEqual(1, len(changes))
        self.assertEqual(arch.FileModification, type(changes[0]))
        self.assertEqual('f2', changes[0].name)
        # commit the other file, relative path
        m['Summary'] = m.description = 'first revision'
        lines = wt.iter_commit(m, file_list=['f1'])
        changes = [L for L in lines if isinstance(L, arch.TreeChange)]
        self.assertEqual(1, len(changes))
        self.assertEqual(arch.FileModification, type(changes[0]))
        self.assertEqual('f1', changes[0].name)
    tests.append('iter_commit_files')


def tla_skip_whitespace_tests(quiet=False):
    from arch.backends import tla
    status = tla.status_cmd(['escape'], expected=(0,1))
    if status == 1 and not quiet:
        sys.stdout.write('[SKIP] ')
        sys.stdout.flush()
    return status == 1


class EscapedWhitespace(TestCase):

    tests = []

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_version()
        self.create_working_tree(self.version)

    def setUp(self):
        TestCase.setUp(self)
        if tla_skip_whitespace_tests(quiet=True): return
        self.tree = self.working_tree
        self.file_name = FileName('foo bar')
        open(self.tree/self.file_name, 'w').write('')
        self.tree.add_tag(self.file_name)
        tagfile = self.tree/'.arch-ids'/self.file_name+'.id'
        print >> open(tagfile, 'w'), 'baby jesus cries'
        self.tree.import_()

    def new_files_escaped(self):
        """New-files from patchlog with spaces"""
        if tla_skip_whitespace_tests(): return
        expected = [self.file_name]
        result = list(self.tree.iter_logs())[0].new_files
        self.assertEqual(expected, result)
    tests.append('new_files_escaped')

    def modified_files_escaped(self):
        """Modified-files from patchlog with spaces"""
        if tla_skip_whitespace_tests(): return
        open(self.tree/self.file_name, 'w').write('Hello, World!')
        m = self.tree.log_message()
        m['Summary'] = 'first revision'
        self.tree.commit(m)
        expected = [self.file_name]
        result = list(self.tree.iter_logs())[1].modified_files
        self.assertEqual(expected, result)
    tests.append('modified_files_escaped')

    def removed_files_escaped(self):
        """Removed-files from patchlog with spaces"""
        if tla_skip_whitespace_tests(): return
        name = self.file_name
        self.tree.delete(name)
        m = self.tree.log_message()
        m['Summary'] = 'first revision'
        self.tree.commit(m)
        expected = [name, name.dirname()/'.arch-ids'/name.basename()+'.id']
        result = list(self.tree.iter_logs())[1].removed_files
        expected.sort()
        result.sort()
        self.assertEqual(expected, result)
    tests.append('removed_files_escaped')

    def get_index_escaped(self):
        """Changeset.get_index supports escaped file names."""
        if tla_skip_whitespace_tests(): return
        open(self.tree/self.file_name, 'w').write('Hello, World!')
        m = self.tree.log_message()
        m['Summary'] = 'first revision'
        self.tree.commit(m)
        cset = self.version['patch-1'].get_patch(self.tree/',cset')
        expected = { 'x_baby_jesus_cries': self.file_name }
        result = cset.get_index('orig-files')
        self.assertEqual(expected, result)
    tests.append('get_index_escaped')

    def file_find_pristine_escaped(self):
        """file_find with pristine supports escaped names."""
        if tla_skip_whitespace_tests(): return
        revision = self.version['base-0']
        result = self.tree.file_find(self.file_name, revision)
        pristine = self.tree.find_pristine(revision)
        expected = pristine/self.file_name
        self.assertEqual(expected, result)
    tests.append('file_find_pristine_escaped')


class LatestVersion(TestCase):

    def extraSetup(self):
        self.set_my_id()

    tests = []

    def latest_version(self):
        return self.version.branch.latest_version()

    def not_registered(self):
        """Branch.latest_version ValueError if archive is not registered."""
        self.assertRaises(ValueError, self.latest_version)
    tests.append('not_registered')

    def missing_branch(self):
        """Branch.latest_version ValueError if branch does not exist."""
        self.create_archive()
        self.assertRaises(ValueError, self.latest_version)
    tests.append('missing_branch')

    def empty_branch(self):
        """Branch.latest_version ValueError if branch is empty."""
        self.create_archive()
        self.create_branch()
        self.assertRaises(ValueError, self.latest_version)
    tests.append('empty_branch')

    def works(self):
        """Branch.latest_version works."""
        self.create_archive()
        self.create_version()
        version = self.version
        branch = version.branch
        self.assertEqual(version, branch.latest_version())
        next_version = arch.Version(version.fullname + '.1')
        next_version.setup()
        self.assertEqual(next_version, branch.latest_version())
    tests.append('works')


class LatestRevision(TestCase):

    def extraSetup(self):
        self.set_my_id()

    tests = []

    def latest_revision(self):
        return self.version.latest_revision()

    def not_registered(self):
        """Version.latest_revision ValueError if archive is not registered."""
        self.assertRaises(ValueError, self.latest_revision)
    tests.append('not_registered')

    def missing_version(self):
        """Version.latest_revision ValueError if version does not exist."""
        self.create_archive()
        self.assertRaises(ValueError, self.latest_revision)
    tests.append('missing_version')

    def empty_version(self):
        """Version.latest_revision ValueError if version is empty."""
        self.create_archive()
        self.create_version()
        self.assertRaises(ValueError, self.latest_revision)
    tests.append('empty_version')

    def works(self):
        """Version.latest_revision works."""
        self.create_archive()
        self.create_version()
        self.create_working_tree(self.version)
        tree = self.working_tree
        tree.import_()
        version = self.version
        self.assertEqual(version['base-0'], version.latest_revision())
        self.commit_summary(tree, 'first revision')
        self.assertEqual(version['patch-1'], version.latest_revision())
    tests.append('works')


class PopulatedArchive(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_version()
        self.create_working_tree(self.version)

    def setUp(self):
        TestCase.setUp(self)
        self.tree = self.working_tree
        self.tree.import_()
        self.commit_summary(self.tree, 'first revision')

    tests = []

    def cachedrevs(self):
        """Cached revisions."""
        def ls_cached():
            gen = self.version.iter_cachedrevs()
            self.assert_(isinstance(gen, types.GeneratorType))
            return list(gen)
        self.assertEqual(ls_cached(), [])
        self.version['base-0'].cache()
        self.assertEqual(ls_cached(), [self.version['base-0']])
        self.version['patch-1'].cache()
        self.assertEqual(ls_cached(),
                         map(lambda(x): self.version[x],
                             ('base-0', 'patch-1')))
        self.version['base-0'].uncache()
        self.assertEqual(ls_cached(), [self.version['patch-1']])
        self.version['patch-1'].uncache()
        self.assertEqual(ls_cached(), [])
    tests.append('cachedrevs')

    def iter_logs(self):
        """SourceTree.iter_log for the tree-version."""
        def ls_logs_revision(vsn=None):
            gen = self.tree.iter_logs(vsn)
            self.assert_(isinstance(gen, types.GeneratorType))
            return map(lambda(x): x.revision, gen)
        rvsns = map(lambda(x): self.version[x], ('base-0', 'patch-1'))
        self.assertEqual(ls_logs_revision(), rvsns)
        self.assertEqual(ls_logs_revision(self.version), rvsns)
        self.assertEqual(ls_logs_revision(self.version.fullname), rvsns)
    tests.append('iter_logs')

    def continuation_from_version(self):
        """Create a continuation version (from version)."""
        version = self.version
        other_version = self.other_version
        assert not other_version.exists()
        version.latest_revision().make_continuation(other_version)
        self.failUnless(other_version.exists())
        revisions = tuple(other_version.iter_revisions())
        other_base = other_version['base-0']
        self.failUnlessEqual(revisions, (other_base,))
        summary = other_base.patchlog.summary
        self.failUnlessEqual(summary, 'tag of %s' % version.latest_revision())
    tests.append('continuation_from_version')

    def mirror_basic(self):
        """Archive.mirror without fromto, pushing."""
        master = self.archive
        mirror_name = master.name + '-MIRROR'
        location = self.arch_dir/mirror_name
        mirror = master.make_mirror(mirror_name, location)
        master.mirror()
        def nonarch(x): return x.nonarch
        master_revisions = map(nonarch, master.iter_revisions())
        mirror_revisions = map(nonarch, mirror.iter_revisions())
        self.failUnlessEqual(master_revisions, mirror_revisions)
    tests.append('mirror_basic')

    def mirror_fromto(self):
        """Archive.mirror with fromto."""
        master = self.archive
        mirror_name = master.name + '-MIRROR'
        mirror = master.make_mirror(mirror_name, self.arch_dir/mirror_name)
        master.mirror()
        third_name = master.name + '-third'
        third = master.make_mirror(third_name, self.arch_dir/third_name)
        master.mirror(fromto=(mirror, third))
        def nonarch(x): return x.nonarch
        mirror_revisions = map(nonarch, mirror.iter_revisions())
        third_revisions = map(nonarch, third.iter_revisions())
        self.failUnlessEqual(mirror_revisions, third_revisions)
    tests.append('mirror_fromto')

    def mirror_limit(self):
        """Archive.mirror with limit version"""
        version = self.version
        other_version = self.other_version
        version.latest_revision().make_continuation(other_version)
        master = self.archive
        mirror_name = master.name + '-MIRROR'
        location = self.arch_dir/mirror_name
        mirror = master.make_mirror(mirror_name, location)
        master.mirror(limit=(version,))
        def nonarch(x): return x.nonarch
        mirror_versions = tuple(map(nonarch, mirror.iter_versions()))
        self.failUnlessEqual(mirror_versions, (version.nonarch,))
    tests.append('mirror_limit')

    def _check_logs(self, tree, levels):
        expected_revs = map(lambda (x): self.version[x], levels)
        found_revs = map(lambda (x): x.revision, tree.iter_logs())
        self.assertEqual(expected_revs, found_revs)

    def tree_get(self):
        """Get a working tree, given a revision."""
        shutil.rmtree(self.working_dir)
        tree = self.version['base-0'].get(self.working_dir)
        self._check_logs(tree, ['base-0'])
    tests.append('tree_get')

    def tree_update(self):
        """WorkingTree.update, without options."""
        shutil.rmtree(self.working_dir)
        tree = self.version['base-0'].get(self.working_dir)
        tree.update()
        self._check_logs(tree, ['base-0', 'patch-1'])
    tests.append('tree_update')


class Merges(PopulatedArchive):

    tests = []

    def setup_first_merge(self):
        version = self.version
        other_version = self.other_version
        tree = self.tree
        version.latest_revision().make_continuation(other_version)
        other_tree = other_version['base-0'].get(self.other_working_dir)
        self.other_tree = other_tree
        self.commit_summary(tree, 'second rev')
        self.commit_summary(tree, 'third rev')

    def star_merge(self):
        """star-merge with a version"""
        self.setup_first_merge()
        other_tree = self.other_tree
        other_tree.star_merge(self.version)
        msg = 'merge'
        self.commit_summary(other_tree, msg)
        log = self.other_version['patch-1'].patchlog
        self.assertEqual(log.summary, msg)
        vsn = self.version
        expected = [vsn['patch-2'], vsn['patch-3']]
        self.assertEqual(log.merged_patches, expected)
    tests.append('star_merge')

    def setup_star_merge(self):
        version, other_version = self.version, self.other_version
        open(self.tree/'foo', 'w').write('hello')
        self.tree.add_tag('foo')
        self.commit_summary(self.tree, 'second rev')
        version.latest_revision().make_continuation(other_version)
        other_tree = other_version['base-0'].get(self.other_working_dir)
        self.other_tree = other_tree
        open(self.tree/'foo', 'w').write('hello world')
        self.commit_summary(self.tree, 'third rev')

    def assertObject(self, obj, class_, **attrs):
        self.assert_(isinstance(obj, class_))
        for k, v in attrs.items():
            self.assertEqual(getattr(obj, k), v)

    def iter_star_merge(self):
        """star-merge incremental output"""
        self.setup_star_merge()
        changes = {}
        for line in self.other_tree.iter_star_merge(self.version):
            if isinstance(line, arch.Chatter):
                continue
            changes[line.name] = line
        expected = (('foo', arch.FileModification, {'binary': False}),
                    ('{arch}/cat/cat--brn/cat--brn--1.0/'
                     'jdoe@example.com--example--9999/patch-log/patch-3',
                     arch.FileAddition, {'directory': False}))
        self.assertEqual(len(expected), len(changes))
        for name, class_, attrs in expected:
            self.assert_(name in changes)
            change = changes[name]
            self.assertObject(change, class_, **attrs)
    tests.append('iter_star_merge')

    def iter_star_merge_conflict(self):
        """incremental star-merge conflict"""
        self.setup_star_merge()
        open(self.tree/'foo', 'w').write('hello world')
        self.commit_summary(self.tree, 'third rev')
        open(self.other_tree/'foo', 'w').write('Hello, World!')
        self.commit_summary(self.other_tree, 'first rev')
        cset_app = self.other_tree.iter_star_merge(self.version)
        for unused in cset_app: pass
        self.assert_(cset_app.conflicting)
    tests.append('iter_star_merge_conflict')

    def setup_merges(self):
        self.setup_first_merge()
        version, other_version = self.version, self.other_version
        tree, other_tree = self.tree, self.other_tree
        other_tree.star_merge(version)
        self.commit_summary(other_tree, 'merge patch-3')
        self.commit_summary(other_tree, 'empty merge')

    def depth_first(self, merges_iter):
        merges = []
        for i, F in merges_iter:
            for f in F:
                merges.append([i, f])
        return merges

    def iter_depth(self, reverse=False):
        """Version.iter_merges depth first"""
        self.setup_merges()
        vsn, oth = self.version, self.other_version
        expected = [[oth['base-0'], vsn['base-0']],
                    [oth['base-0'], vsn['patch-1']],
                    [oth['base-0'], oth['base-0']],
                    [oth['patch-1'], vsn['patch-2']],
                    [oth['patch-1'], vsn['patch-3']],
                    [oth['patch-1'], oth['patch-1']],
                    [oth['patch-2'], oth['patch-2']]]
        if reverse: expected.reverse()
        merges_iter = self.other_version.iter_merges(reverse=reverse)
        merges = self.depth_first(merges_iter)
        self.assertEqual(expected, merges)
    tests.append('iter_depth')

    def iter_depth_reverse(self):
        """Version.iter_merges depth first in reverse"""
        self.iter_depth(reverse=True)
    tests.append('iter_depth_reverse')

    def iter_not_me(self):
        """Version.iter_merges metoo=False"""
        self.setup_merges()
        vsn, oth = self.version, self.other_version
        expected = [[oth['base-0'], vsn['base-0']],
                    [oth['base-0'], vsn['patch-1']],
                    [oth['patch-1'], vsn['patch-2']],
                    [oth['patch-1'], vsn['patch-3']]]
        merges = self.depth_first(self.other_version.iter_merges(metoo=False))
        self.assertEqual(expected, merges)
    tests.append('iter_not_me')

    def iter_version(self):
        """Version.iter_merges limited to a single source version"""
        self.setup_merges()
        vsn, oth = self.version, self.other_version
        expected = [[oth['base-0'], oth['base-0']],
                    [oth['patch-1'], oth['patch-1']],
                    [oth['patch-2'], oth['patch-2']]]
        merges_iter = self.other_version.iter_merges(self.other_version)
        merges = self.depth_first(merges_iter)
        self.assertEqual(expected, merges)
    tests.append('iter_version')

    def breadth_first(self, iterator):
        merges = map(list, iterator)
        for step in merges:
            step[1] = list(step[1])
        return merges

    def iter_breadth(self):
        """Version.iter_merges with breadth first traversal."""
        self.setup_merges()
        merges_iter = self.other_version.iter_merges()
        merges = self.breadth_first(merges_iter)
        vsn, oth = self.version, self.other_version
        expected = [[oth['base-0'], [vsn['base-0'],
                                     vsn['patch-1'],
                                     oth['base-0']]],
                    [oth['patch-1'], [vsn['patch-2'],
                                      vsn['patch-3'],
                                      oth['patch-1']]],
                    [oth['patch-2'], [oth['patch-2']]]]
        self.assertEqual(expected, merges)
    tests.append('iter_breadth')


class RevisionFiles(PopulatedArchive):

    tests = []

    def parse_checksum(self):
        """Parsing checksum file data"""
        checksum = '\n'.join((
            "-----BEGIN PGP SIGNED MESSAGE-----",
            "Hash: SHA1",
            "",
            "Signature-for: ddaa@ddaa.net--2004/pyarch--ddaa--0.5--base-0",
            "md5 pyarch--ddaa--0.5--base-0.tar.gz"
            " f1998e8294e50070993471cec6b66b52",
            "foo bar-baz-blah 0123456789abcdef",
            "sha1 pyarch--ddaa--0.5--base-0.tar.gz"
            " 633c0efcceab4e6588a04a9ce73233bc206bcdf9",
            "-----BEGIN PGP SIGNATURE-----",
            "Version: GnuPG v1.2.4 (GNU/Linux)",
            "",
            "iD8DBQFAUQ/bZEH9AkgfRL0RAiiFAKCnV9jenFvCsckbIigtQmHYfZgcKwCgk5FY",
            "XQwvigYzMiiR9lFZh6vYUJM=",
            "=PCU/",
            "-----END PGP SIGNATURE-----",
            ""))
        rvsn  = arch.Revision('ddaa@ddaa.net--2004/pyarch--ddaa--0.5--base-0')
        checksums = rvsn._parse_checksum(checksum)
        expected = {'pyarch--ddaa--0.5--base-0.tar.gz':
                    {'md5': 'f1998e8294e50070993471cec6b66b52',
                     'sha1': '633c0efcceab4e6588a04a9ce73233bc206bcdf9'},
                    'bar-baz-blah': {'foo': '0123456789abcdef'}}
        self.assertEqual(expected, checksums)
    tests.append('parse_checksum')

    def _help_file_names(self, rvsn, expected_names):
        files = rvsn.iter_files()
        file_names = map(lambda (x): x.name, files)
        file_names.sort()
        expected_names.sort()
        self.assertEqual(file_names, expected_names)

    def import_files(self):
        """Import revisions have expected files."""
        rvsn  = self.version['base-0']
        expected_names = ['checksum', 'log', rvsn.nonarch + '.src.tar.gz' ]
        self._help_file_names(rvsn, expected_names)
    tests.append('import_files')

    def commit_files(self):
        """Commit revisions (uncached) have expected files."""
        rvsn = self.version['patch-1']
        expected_names = ['checksum', 'log', rvsn.nonarch + '.patches.tar.gz']
        self._help_file_names(rvsn, expected_names)
    tests.append('commit_files')

    def cachedrev_files(self):
        """Commit revisions (cached) have expected files."""
        rvsn = self.version['patch-1']
        rvsn.cache()
        expected_names = ['checksum', 'log', rvsn.nonarch + '.patches.tar.gz',
                          'checksum.cacherev', rvsn.nonarch + '.tar.gz']
        self._help_file_names(rvsn, expected_names)
    tests.append('cachedrev_files')

    def revision_files_data(self):
        """Accessing the data of Revision files."""
        rvsn = self.version['patch-1']
        rvsn.cache()
        rev_dir = reduce(
            lambda a,b: a/DirName(b), \
            (self.arch_dir, self.arch_name,
             'cat', 'cat--brn', 'cat--brn--1.0', 'patch-1'))
        for rev_file in rvsn.iter_files():
            self.assert_(rev_file.data == open(rev_dir/rev_file.name).read())
    tests.append('revision_files_data')


class PristineTrees(PopulatedArchive):

    tests = []

    def iter_pristines_one(self):
        """iter_pristines works"""
        result = list(self.tree.iter_pristines())
        expected = [self.version['patch-1']]
        self.assertEqual(expected, result)
    tests.append('iter_pristines_one')

    def add_pristine(self):
        """add_pristine works"""
        self.tree.add_pristine(self.version['base-0'])
        result = list(self.tree.iter_pristines())
        expected = [self.version['base-0'], self.version['patch-1']]
        self.assertEqual(expected, result)
    tests.append('add_pristine')

    def find_pristine(self):
        """find_pristine works"""
        revision = self.version['patch-1']
        result = self.tree.find_pristine(revision)
        expected = (self.tree / '{arch}' / '++pristine-trees' / 'unlocked' /
                    self.version.category.nonarch /
                    self.version.branch.nonarch / self.version.nonarch /
                    self.archive.name / revision.nonarch)
        self.assertEqual(expected, result)
        self.assertEqual(arch.ArchSourceTree, type(result))
    tests.append('find_pristine')

    def find_pristine_missing(self):
        """find_pristine raises NoPristineFoundError"""
        revision = self.version['base-0']
        self.assert_(revision not in self.tree.iter_pristines())
        def thunk(): return self.tree.find_pristine(revision)
        self.assertRaises(arch.errors.NoPristineFoundError, thunk)
    tests.append('find_pristine_missing')


class Changeset(PopulatedArchive):

    tests = []

    def setUp(self):
        PopulatedArchive.setUp(self)
        self.base0 = self.version['base-0']
        self.patch1 = self.version['patch-1']
        self.base0_tree = self.base0.get(self.arch_dir/'base0')
        self.patch1_tree = self.patch1.get(self.arch_dir/'patch1')
        self.changeset_name = self.arch_dir/'cset'

    def changeset_does_nothing(self):
        """Changeset.does_nothing works."""
        cset = arch.delta(
            self.patch1_tree, self.patch1_tree, self.changeset_name)
        self.assertEqual(True, cset.does_nothing)
        shutil.rmtree(cset)
        cset = arch.delta(
            self.base0_tree, self.patch1_tree, self.changeset_name)
        self.assertEqual(False, cset.does_nothing)
    tests.append('changeset_does_nothing')


class Delta(Changeset):

    tests = []

    def _patchlog_name(self, rvsn):
        return '{arch}/%s/%s/%s/%s/patch-log/%s' % \
               (rvsn.category.nonarch, rvsn.branch.nonarch,
                rvsn.version.nonarch, rvsn.archive.name, rvsn.patchlevel)

    def _delta_helper(self, orig, mod):
        name = self._patchlog_name(self.patch1)
        expected_created = [('A_./'+name, name)]
        cset = arch.delta(orig, mod, self.changeset_name)
        created = list(cset.iter_created_files(True))
        self.assertEqual(expected_created, created)

    def delta_tree_tree(self):
        """delta(tree, tree) works."""
        self._delta_helper(self.base0_tree, self.patch1_tree)
    tests.append('delta_tree_tree')

    def delta_tree_revision(self):
        """delta(tree, revision) works."""
        self._delta_helper(self.base0_tree, self.patch1)
    tests.append('delta_tree_revision')

    def delta_revision_tree(self):
        """delta(revision, tree) works."""
        self._delta_helper(self.base0, self.patch1_tree)
    tests.append('delta_revision_tree')

    def delta_revision_revision(self):
        """delta(revision, revision) works."""
        self._delta_helper(self.base0, self.patch1)
    tests.append('delta_revision_revision')

    def iter_delta(self):
        """iter_delta works."""
        orig, mod = self.base0, self.patch1
        name = self._patchlog_name(self.patch1)
        idelta = arch.iter_delta(orig, mod, self.changeset_name)
        def get_changeset(): return idelta.changeset
        self.assertRaises(AttributeError, get_changeset)
        lines = [ L for L in idelta if not isinstance(L, arch.Chatter)]
        self.assertEqual(1, len(lines))
        self.assertEqual(arch.FileAddition, type(lines[0]))
        self.assertEqual(name, lines[0].name)
        cset = get_changeset()
        created = list(cset.iter_created_files(True))
        expected_created = [('A_./'+name, name)]
        self.assertEqual(expected_created, created)
    tests.append('iter_delta')


class TreeChanges(Delta):
    tests = []

    def changes(self):
        """WorkingTree.changes works"""
        cset = self.tree.changes(None, self.changeset_name)
        self.assertEqual(True, cset.does_nothing)
        shutil.rmtree(cset)
        cset = self.tree.changes(self.base0, self.changeset_name)
        log_name = self._patchlog_name(self.patch1)
        expected = [('A_./%s' % log_name, log_name)]
        newfiles = list(cset.iter_created_files(True))
        self.assertEqual(expected, newfiles)
        shutil.rmtree(cset)
        open(self.tree/'foo', 'w').write("hello\n")
        self.tree.add_tag('foo')
        cset = self.tree.changes(None, self.changeset_name)
        expected = ['.arch-ids/foo.id', 'foo']
        newfiles = [C[1] for C in cset.iter_created_files(True)]
        newfiles.sort()
        self.assertEqual(expected, newfiles)
    tests.append('changes')


class FileHistory(TestCase):

    file_history = [
        { '%add': ['foo'], 'foo': "hello world\n" },
        { 'foo': "Hello, World\n" } ]

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_version()
        self.create_working_tree(self.version)

    def setUp(self):
        TestCase.setUp(self)
        self.tree = self.working_tree
        self.make_history(self.tree, self.file_history)


class Patchlog(TestCase):

    def extraSetup(self):
        self.set_my_id()
        self.create_archive()
        self.create_version()
        self.create_working_tree(self.version)

    def setUp(self):
        TestCase.setUp(self)
        self.tree = self.working_tree

    tests = []

    def renames(self):
        """Patchlog.renamed_files works."""
        file_history = [
            { '%add': ['foo', 'bar'], 'foo': "hello\n", 'bar': "world\n" },
            { '%mv': [['foo', 'spam'], ['bar', 'eggs']] } ]
        self.make_history(self.tree, file_history)
        log = list(self.tree.iter_logs())[1]
        def filetag(name):
            name = FileName(name)
            return name.dirname()/'.arch-ids'/name.basename()+'.id'
        expected = {'foo':'spam', filetag('foo'): filetag('spam'),
                    'bar':'eggs', filetag('bar'): filetag('eggs')}
        self.assertEqual(expected, log.renamed_files)
    tests.append('renames')

    def get_missing(self):
        """Patchlog getitem on a missing header returns None."""
        self.tree.import_()
        log = self.tree.tree_revision.patchlog
        self.assertEqual(None, log['Not-Present-Header'])
    tests.append('get_missing')

    def revision(self):
        """Patchlog.revision works."""
        self.tree.import_()
        rvsn = self.tree.tree_revision
        self.assertEqual(rvsn, rvsn.patchlog.revision)
    tests.append('revision')

    def summary(self):
        """Patchlog.summary works."""
        self.tree.import_()
        rvsn = self.tree.tree_revision
        self.assertEqual("initial import", rvsn.patchlog.summary)
    tests.append('summary')


class FileFind(FileHistory):

    tests = []

    def file_find_pristine(self):
        """file_find works with pristines"""
        revision = self.version['patch-1']
        path = self.tree.file_find('foo', revision)
        self.assertEqual(PathName, type(path))
        pristine = self.tree.find_pristine(revision)
        expected = pristine/'foo'
        self.assertEqual(expected, path)
        result = open(path).read()
        expected = self.file_history[1]['foo']
        self.assertEqual(expected, result)
    tests.append('file_find_pristine')

    def file_find_revlib(self):
        """file_find works with the revision library"""
        revlib_dir = self.arch_dir/'revlib'
        os.mkdir(revlib_dir)
        arch.register_revision_library(revlib_dir)
        revision = self.version['patch-1']
        shutil.rmtree(self.tree/'{arch}'/'++pristine-trees')
        self.assertEqual([], list(self.tree.iter_pristines()))
        revision.library_add()
        path = self.tree.file_find('foo', revision)
        self.assertEqual(PathName, type(path))
        pristine = revision.library_find()
        expected = pristine/'foo'
        self.assertEqual(expected, path)
        result = open(path).read()
        expected = self.file_history[1]['foo']
        self.assertEqual(expected, result)
    tests.append('file_find_revlib')

    def file_find_pristine_missing(self):
        """file_find works with missing pristine"""
        revision = self.version['base-0']
        path = self.tree.file_find('foo', revision)
        self.assertEqual(PathName, type(path))
        pristine = self.tree.find_pristine(revision)
        expected = pristine/'foo'
        self.assertEqual(expected, path)
        result = open(path).read()
        expected = self.file_history[0]['foo']
        self.assertEqual(expected, result)
    tests.append('file_find_pristine_missing')

    def _file_find_raises_helper(self, flavour):
        revision = self.version['patch-1']
        name = {'source': 'bar', 'junk': ',bar'}[flavour]
        open(self.tree/name, 'w').write('Yadda!\n')
        if flavour == 'source': self.tree.add_tag(name)
        def thunk(): return self.tree.file_find(name, revision)
        self.assertRaises(arch.errors.MissingFileError, thunk)

    def file_find_new_raises(self):
        """file_find raises MissingFileError for new source files"""
        self._file_find_raises_helper('source')
    tests.append('file_find_new_raises')

    def file_find_junk_raises(self):
        """file_find raises MissingFileError for junk files"""
        self._file_find_raises_helper('junk')
    tests.append('file_find_junk_raises')


class ApplyChangeset(FileHistory, Delta):

    tests = []

    def setUp(self):
        FileHistory.setUp(self)
        shutil.rmtree(self.working_dir)
        self.tree = None
        self.base0 = self.version['base-0']
        self.patch1 = self.version['patch-1']
        self.changeset_name = self.arch_dir/'cset'
        self.cset = arch.delta(self.base0, self.patch1, self.changeset_name)

    def iter_apply(self):
        """Changeset.iter_apply works"""
        tree = self.base0.get(self.working_dir)
        lines = list(self.cset.iter_apply(tree))
        self.assertEqual(2, len(lines))
        self.assertEqual(arch.FileAddition, type(lines[0]))
        self.assertEqual(self._patchlog_name(self.patch1), lines[0].name)
        self.assertEqual(arch.FileModification, type(lines[1]))
        self.assertEqual('foo', lines[1].name)
        self.assertEqual(self.patch1, list(tree.iter_logs())[-1].revision)
        self.assert_(not tree.has_changes())
    tests.append('iter_apply')

    def iter_apply_conflict(self):
        """Changeset.iter_apply works with conflicts."""
        tree = self.patch1.get(self.working_dir)
        lines = list(self.cset.iter_apply(tree))
        self.assertEqual(2, len(lines))
        self.assertEqual(arch.FileAddition, type(lines[0]))
        self.assertEqual(self._patchlog_name(self.patch1), lines[0].name)
        self.assertEqual(arch.PatchConflict, type(lines[1]))
        self.assertEqual('foo', lines[1].name)
        self.assertEqual(self.patch1, list(tree.iter_logs())[-1].revision)
        self.assert_(not tree.has_changes())
    tests.append('iter_apply_conflict')

    def iter_apply_reverse(self):
        """Changeset.iter_apply reverse=True works."""
        tree = self.patch1.get(self.working_dir)
        lines = list(self.cset.iter_apply(tree, reverse=True))
        self.assertEqual(arch.FileDeletion, type(lines[0]))
        self.assertEqual(self._patchlog_name(self.patch1), lines[0].name)
        self.assertEqual(arch.FileModification, type(lines[1]))
        self.assertEqual('foo', lines[1].name)
        self.assertEqual(self.base0, list(tree.iter_logs())[-1].revision)
        self.assert_(not tree.has_changes())
    tests.append('iter_apply_reverse')


def test_suite(limit=()):
    classes = (
        'ArchBasic',
        'Archive',
        'MirrorArchive',
        'InitTree',
        'InventoryTree',
        'WorkingTree',
        'TreeRevision',
        'LogMessage',
        'LogMessageWhitespace',
        'Import',
        'Commit',
        'EscapedWhitespace',
        'LatestVersion',
        'LatestRevision',
        'PopulatedArchive',
        'Merges',
        'RevisionFiles',
        'PristineTrees',
        'Changeset',
        'Delta',
        'TreeChanges',
        'Patchlog',
        'FileFind',
        'ApplyChangeset',
        )
    return framework.make_test_suite(globals(), classes, limit)

def main(argv):
    return framework.run_test_suite(argv, test_suite)

if __name__ == "__main__":
    sys.exit(main(sys.argv))
