# Copyright (C) 2004 Aaron Bentley
# <aaron.bentley@utoronto.ca>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
import os

#ensure that the parent directory is in the path (for epydoc)
sys.path=[os.path.realpath(os.path.dirname(__file__)+"/..")]+sys.path

import commands
import cmdutil
import arch
import pylon
import time
import terminal

__docformat__ = "restructuredtext"
__doc__ = "command template"

class Annotate(commands.BaseCommand):
    """Print an annotated version of a file"""
    def __init__(self):
        self.description = self.__doc__

#   override get_completer if you want custom completion
#    def get_completer(self, arg, index):
#        return cmdutil.iter_dir_completions(arg)

    def do_command(self, cmdargs):
        parser=self.get_parser()
        (options, args) = parser.parse_args(cmdargs)

        if len(args) < 1 or len(args) > 2:
            raise pylon.errors.GetHelp
        file = args[0]
        tree = arch.tree_root(".")
        if len(args) == 2:
            revision = cmdutil.determine_revision_tree(tree, args[1])
        else:
            revision = pylon.tree_latest(tree)
        try:
            lines = pylon.annotate_file(tree, revision, file)
        except IOError, e:
            raise pylon.errors.CommandFailedWrapper(e)
        if options.deep:
            diffs = {}
            for line in lines:
                if line.log is None:
                    continue
                line.log = pylon.merge_blame(line.log, file, 
                                             pylon.tree_cwd(tree)+"/"+file, 
                                             line.text, diffs)
        if options.baz_style or options.patchlevels:
            pre_tag = True
            ancestors = list(pylon.iter_ancestry(tree))
            recent_ancestors = []
            name_map = {}
            
            name_map["UNKNOWN"] = str(ancestors[-1])
            if (options.patchlevels):
                anc_level = 0
                for revision in ancestors:
                    if anc_level == 0 and revision.version != tree.tree_version:
                        anc_level = 1
                    if anc_level > 0:
                        name_map[str(revision)] = "ances-%i" % anc_level
                        anc_level += 1
                    else:
                        name_map[str(revision)] = revision.patchlevel
            baz_display(lines, name_map)
        else:
            colorized_display(lines, options)
        return 0

    def get_parser(self):
        parser = cmdutil.CmdOptionParser("annotate FILE [REVISION]")
        parser.add_option("-s", "--summary", action="store_true", 
                          dest="summary", help="Show patchlog summary")
        parser.add_option("-D", "--date", action="store_true", 
                          dest="date", help="Show patchlog date")
        parser.add_option("-c", "--creator", action="store_true", 
                          dest="creator", help="Show the id that committed the"
                          " revision")
        parser.add_option("-b", "--baz", action="store_true", dest="baz_style", 
                          help="Show annotations in baz style")
        parser.add_option("-p", "--patchlevels", action="store_true", 
                          dest="patchlevels", 
                          help="Show patchlevels (and ances for ancestor"
                          " revisions)")
        parser.add_option("--notws", action="store_true", 
                          dest="notws", 
                          help="Show, but don't annotate whitespace lines")
        parser.add_option("", "--deep", action="store_true", 
                          dest="deep", 
                          help="Assign blame to merged revisions")
        return parser

    def help(self, parser=None):
        """
        Prints a help message.

        :param parser: If supplied, the parser to use for generating help.  If \
        not supplied, it is retrieved.
        :type parser: cmdutil.CmdOptionParser
        """
        if parser==None:
            parser=self.get_parser()
        parser.print_help()
        print """
Annotate will print up a version of a file, showing which revision added
each line.

Lines whose derivation cannot be determined as listed as UNASSIGNED.

The --deep option attempts to determine which merge inserted a line, but
because merges are inexact patching, it can be confused by lines which are
repeated more than once in a revision's diff (e.g. whitespace).
"""
def baz_display(lines, name_map={}):
    for line in lines:
        if line.log == None:
            revision_text = "UNKNOWN"
        else:
            revision_text = str(line.log.revision)
        try:
            while(True):
                revision_text = name_map[revision_text]
        except KeyError, e:
            pass
        print "%s: %s" % (revision_text, line.text.rstrip("\n"))


def colorized_display(lines, options):
    last = None
    for line in lines:
        if last is None or str(line.log) != str(last.log):
            if options.notws and line.text.strip() == "":
                continue
            if line.log == None:
                print terminal.colorstring("UNASSIGNED", 'blue')
            else:
                print terminal.colorstring(str(line.log.revision), 'blue')
                if options.creator:
                    print terminal.colorstring("    %s" % line.log.creator,
                                               'yellow')
                if options.date:
                    print terminal.colorstring("    %s" % \
                        time.strftime('%Y-%m-%d %H:%M:%S %Z',
                        line.log.date), 'yellow')
                if options.summary and line.log.summary != "":
                    print terminal.colorstring("    %s" % line.log.summary,
                                               'yellow')
            last=line
        print(line.text.rstrip("\n"))


#This function assigns the command class to a user command name
def add_command(commands):
    commands["annotate"] = Annotate



# arch-tag: c9833e6f-0ebe-4cb9-a799-d9ec2e50674e
